/* --------------------------------------------------------------- */
/* DOGXL.c -- DOGXL LCD driver                                     */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This module provides the glue between tollosDraw and a DOGXL    */
/* LCD screen, including a screen buffer.                          */
/*                                                                 */
/* Measured performance, etc.:                                     */
/* * Time to write all pels using drawSetPelIO: 7.7ms (96 MHz MCU) */
/* * Time to send whole buffer to DOGXL: 5.3 ms (12 MHz SPI)       */
/* * Power consumption: 1.04 mA @ 3.3V (160L-7, all pels full on). */
/*                                                                 */
/* This glue assumes the DOGXL is attached to a microcontroller in */
/* the 3-wire 9-bit SPI configuration, using the primary SPI port  */
/* and associated select line (as defined by the spi9 glue         */
/* functions), and with the SPI clock set to 12 MHz.               */
/*                                                                 */
/* This 3-wire 9-bit SPI combination is about 33x faster than      */
/* 400 kHz I2C and theoretically could write the 4160-byte screen  */
/* buffer in 3.12 ms, running the bus clock at 12 MHz.             */
/* In practice, there is some inter-byte overhead and the          */
/* measured time is 5.3 ms.                                        */
/*                                                                 */
/* At this speed it is unlikely to be valuable to double-buffer    */
/* the screen in order to only send changes, because the           */
/* inter-byte overhead is already significant.                     */
/*                                                                 */
/* The screen buffer for DOGXL is maintained in the order that is  */
/* is required for sending to the display; this is determined by   */
/* the UC1610 controller on the display.                           */
/*                                                                 */
/* Each byte stored covers one pel in each of four rows in a       */
/* column of four pels (4 rows together are called a 'page' in     */
/* the UC1610 datasheet).  The byte in which a given pel is to be  */
/* found is therefore given by ((DOGY-1-y)/4)*160+x, with the two  */
/* bits being determined by y%4.  Note that the two bits for the   */
/* top row go into the two least significant bits of the byte.     */
/*                                                                 */
/* The "lookup tables" pelshift and pelmasks are used to speed up  */
/* the retrieval and setting of a pel within a byte.               */
/*                                                                 */
/* X and Y coordinates are always checked so there is no           */
/* possibility of buffer overrun or addressing exception.          */
/*                                                                 */
/* Pel data values are not checked on set (they may cause local    */
/* pel corruption but will not cause addressing errors).           */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollos.h"
#include "DOGXL160.h"              // UC1610 definitions

// the screen buffer; this is mapped to match the DOGXL page
// organization (see above)
static byte buff[DOGBYTES];        // screen buffer

// constants
static const byte pelshift[4]={   0,    2,    4,    6};
static const byte pelmasks[4]={0xfc, 0xf3, 0xcf, 0x3f};

/* --------------------------------------------------------------- */
/* drawClearIO -- set every pel to the same value                  */
/*                                                                 */
/*   pel -- new value for the screen                               */
/*                                                                 */
/* This sets every pel in the screen buffer to the provided value. */
/* --------------------------------------------------------------- */
void drawClearIO(int pel) {
  byte *bp;                   // work
  uint val;                   // ..

  val=pel&0x03;               // just 2 bits
  val=val | val<<2 | val<<4 | val<<6;

  for (bp=buff; bp<buff+DOGBYTES; bp++) *bp=val;
  } // drawClearIO

/* --------------------------------------------------------------- */
/* drawFromRGBIO -- convert RGB to a pel value                     */
/*                                                                 */
/*   rgb -- colour value [0..255 in each]                          */
/*                                                                 */
/*   returns the corresponding pel value                           */
/*                                                                 */
/* This converts standard RGB to a private pel value suitable for  */
/* use in drawSetPelIO, etc.                                       */
/* --------------------------------------------------------------- */
int drawFromRGBIO(drawRGB rgb) {
  uint v;                          // value

  // This uses the common v = 0.3*r + 0.59*g + 0.11*b, slightly
  // modified so that division will be by a power of two:
  //   v = (77*r + 151*g + 28*b)/256
  // The grey level value v is then mapped to the nearest of the
  // four DOGXL values (in practice, without dividing by 256 first).

  v=(uint)rgb.r*77 + (uint)rgb.g*151 + (uint)rgb.b*28;
  // now in 0..65280

  // round to nearest of 0..3, keeping power of 2 divisor
  v=(v+10880)*3 / (256*256);

  return 3-v;                      // invert: 00=white, 11=black
  } // drawFromRGBIO

/* --------------------------------------------------------------- */
/* drawGetPelIO -- get pel value                                   */
/*                                                                 */
/*   x   -- pel X coordinate                                       */
/*   y   -- pel Y coordinate                                       */
/*                                                                 */
/*   returns the value of the pel at x, y                          */
/*                                                                 */
/* If x or y is too large or less than zero, 0 is returned.        */
/* --------------------------------------------------------------- */
int drawGetPelIO(int x, int y) {
  byte b;                          // work
  if ((uint)x>=DOGX || (uint)y>=DOGY) return 0;

  y=DOGY-1-y;                      // DOGXL top-down row number

  b=buff[(y/4)*DOGPAGEX+x];        // byte that has the pel
  b>>=pelshift[y%4];               // shift desired pel to right
  return b & 0x03;                 // and return 2 bits
  } // drawGetPelIO

/* --------------------------------------------------------------- */
/* drawInitIO -- initialize display and discover its configuration */
/*                                                                 */
/*   config -- configuration structure (always set)                */
/*                                                                 */
/*   returns 0 if OK, negative if error (e.g., device not          */
/*     available, or bus error)                                    */
/*                                                                 */
/* This function initializes the display, clears the screen buffer */
/* to white, and sets the configuration structure so the caller    */
/* can determine the device geometry and characteristics.  The     */
/* screen buffer is cleared and the configuration is set even if   */
/* an error occurred during hardware initialization.               */
/*                                                                 */
/* The DOGXL device is write only when SPI-connected, so this      */
/* function always returns 0 unless there is an SPI bus error.     */
/* --------------------------------------------------------------- */
int drawInitIO(drawConfig *config) {
  int rc=0;                   // work

  // set the return data (all values must be < 2**31)
  config->type=DOGTYPE;
  config->width=DOGX;
  config->height=DOGY;
  config->depth=DOGZ;
  config->xpitch=DOGXPITCH;
  config->ypitch=DOGYPITCH;
  config->writetime=DOGTIME;

  // clear buffer to white
  drawClearIO(0);

  // initialize the SPI device hardware
  spi9InitIO();

  // set up UC1610 controller (accumulate rc through the commands)

  rc|=spi9CommandIO(UC_Reset);               // reset the controller

  rc|=spi9Command2IO(UC_COMend, DOGY-1);     // last live COM electrode

  rc|=spi9CommandIO(UC_Mapping);             // default (6 o'clock) mapping

  rc|=spi9CommandIO(UC_PanelLoad | 0x3);     // panel loading 28..38nF

  rc|=spi9CommandIO(UC_BiasRatio | 0x3);     // bias ratio to 12

  rc|=spi9Command2IO(UC_Vbias, 90);          // set contrast = 90 / 255
                                             // max 141 when bias ratio is 12
  rc|=spi9CommandIO(UC_AddressControl | 1);  // auto-wrap, column first

  rc|=spi9CommandIO(UC_GrayShade | 0);       // 24% gray spread

  // if any of the above failed, do not enable
  if (rc<0) {
    //debugf("*** bad rc during DOGXL init\n");
    return -1;
    }

  rc|=spi9CommandIO(UC_Enable | 1);          // enable display

  if (rc<0) {
    //debugf("*** bad rc during DOGXL enable\n");
    return -2;
    }

  // test code to add a short line from top left for initial test
  //   for (rc=0; rc<5; rc++) buff[rc*DOGX+rc]=0xff;

  return drawWriteIO();             // write cleared buffer to device
  } // drawInitIO

/* --------------------------------------------------------------- */
/* drawRawIO -- read and/or write the screen buffer 'as is'        */
/*                                                                 */
/*   to   -- first byte to copy screen buffer to, or NULL for no   */
/*           read                                                  */
/*   from -- first byte to copy screen buffer to, or NULL for no   */
/*           write                                                 */
/*                                                                 */
/* This function is used to provide private access to the screen   */
/* buffer, which will be in device-specific format.  The entire    */
/* buffer is copied, with a read (copy to the caller) taking       */
/* place before the write (copy to the screen buffer)              */
/*                                                                 */
/* This is intended for experimental use or when performance is    */
/* critical.  Tollos code rules do not permit direct cross-module  */
/* memory sharing, so a copy is still necessary.                   */
/* --------------------------------------------------------------- */
void drawRawIO(byte *to, const byte *from) {
  if (to!=NULL) {
    byte *bf=buff;
    for  (; bf<buff+DOGBYTES; bf++, to++) *to=*bf;
    }
  if (from!=NULL) {
    byte *bt=buff;
    for  (; bt<buff+DOGBYTES; bt++, from++) *bt=*from;
    }
  } // drawRawIO

/* --------------------------------------------------------------- */
/* drawSetPelIO -- set pel to a new value                          */
/*                                                                 */
/*   x   -- pel X coordinate                                       */
/*   y   -- pel Y coordinate                                       */
/*   pel -- new value for the pel                                  */
/*                                                                 */
/* If x or y is too large or less than zero, this function has     */
/* no effect.                                                      */
/* If pel is too large some pels in the buffer may be undefined.   */
/*                                                                 */
/* The code below runs in about 1.85 us on a 96 MHz MCU, so to     */
/* write every pel on the screen individually takes about 7.7ms.   */
/* --------------------------------------------------------------- */
void drawSetPelIO(int x, int y, int pel) {
  byte b, *bp;                     // work
  if ((uint)x>=DOGX || (uint)y>=DOGY) return;

  y=DOGY-1-y;                      // DOGXL top-down row number
  b=pel<<pelshift[y%4];            // set up pel bits
  bp=buff+(y/4)*DOGPAGEX+x;        // -> byte that has the pel
  *bp=(*bp&pelmasks[y%4]) | b;     // .. and insert into byte
  } // drawSetPelIO

/* --------------------------------------------------------------- */
/* drawStopIO -- stop the display device                           */
/*                                                                 */
/* This puts the device into a clean or low power/off mode.  For   */
/* LCDs this should save power as well as properly discharging the */
/* display.                                                        */
/* --------------------------------------------------------------- */
void drawStopIO(void) {
  spi9CommandIO(UC_Reset);         // reset (low power) mode, please
  } // drawStopIO

/* --------------------------------------------------------------- */
/* drawToRGBIO -- convert pel value to RGB                         */
/*                                                                 */
/*   pel -- pel value to convert                                   */
/*   rgb -- returned colour value [0..255 in each]                 */
/*                                                                 */
/* This converts a private pel value back to standard RGB.         */
/* --------------------------------------------------------------- */
void drawToRGBIO(int pel, drawRGB *rgb) {
  byte w;                     // work

  w=3-(pel & 0x3);            // clean and make 11 black
  w*=85;                      // scale to 0..255
  rgb->r=w;                   // return the three values
  rgb->g=w;                   // ..
  rgb->b=w;                   // ..
  } // drawToRGBIO

/* --------------------------------------------------------------- */
/* drawWriteIO -- write buffer to screen (draw)                    */
/*                                                                 */
/*   returns 0 if OK, negative if error (e.g., bus error)          */
/*                                                                 */
/* This function sends the screen buffer to the device.            */
/*                                                                 */
/* For the DOGXL, this is done by sending a 'page' at a time,      */
/* where a page is four rows of pels, starting at the top of the   */
/* screen.  The bytes to be sent are a vertical swathe of the      */
/* pels, 2 bits per row.                                           */
/*                                                                 */
/* Later a dirty flag per page might be added, so that only        */
/* changed pages will be sent.                                     */
/*                                                                 */
/* The entire screen buffer could be sent as a single write; the   */
/* send by page has the advantage that if a noise burst corrupts   */
/* a transmission then only one page will be corrupt instead of    */
/* perhaps the entire screen.                                      */
/*                                                                 */
/* The time to write all 26 pages to the DOGXL using this code and */
/* an SPI clock of 12 MHz (the maximum, using the 3-wire 9-bit     */
/* connection) is 5.3 ms (800,000 9-bit 'bytes' per second).       */
/* --------------------------------------------------------------- */
int drawWriteIO(void) {
  uint page;                  // 'page' number
  byte *bp;                   // data pointer
  int rc=0;                   // return code

  // a page has four rows of 2-bit pels (which is 1 byte/column)

  for (page=0, bp=buff; page<(DOGY/4) && rc==0; page++, bp+=DOGX) {

    // send X and Y
    rc|=spi9CommandIO(UC_ColumnAddrMSB | 0); // X=0 ..
    rc|=spi9CommandIO(UC_ColumnAddrLSB | 0); // ..
    rc|=spi9CommandIO(UC_PageAddr | page);   // Y=page

    // send the data for the page
    rc|=spi9DataIO(bp, DOGPAGEX);
    } // page

  return rc;
  } // drawWriteIO

