/* --------------------------------------------------------------- */
/* lpcSysIO.c -- system-dependent core functions                   */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These are processor-specific functions for Tollos.              */
/*                                                                 */
/* They should not normally be used directly by applications.      */
/* --------------------------------------------------------------- */

#define  USEPLL 1             // set = 0 to use IRC clock for the
                              // system clock, e.g., for testing

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"
#include "LPC17xx.h"
#include "lpc17xx_clkpwr.h"   // for power settings
#include "lpc17xx_systick.h"  // for SysTick
#include "lpc17xx_wdt.h"      // for watchdog
#include "system_LPC17xx.h"   // for SystemInit

// Local statics
static int  systick=0;        // SysTick interval us; 0 until first set
static uint watchtime=0;      // Watchdog us; 0 until first set
static void stopPLL0(void);   // turn off PLL0 as per errata

/* --------------------------------------------------------------- */
/* sysFeedDogIO -- feed the watchdog (if started)                  */
/*                                                                 */
/* Once the watchdog has been started, this function is called to  */
/* reset its counter and prevent a Watchdog Reset.  If called      */
/* before the watchdog is started it has no effect.                */
/* --------------------------------------------------------------- */
void sysFeedDogIO(void) {
  if (watchtime!=0) WDT_Feed();
  } // sysFeedDogIO

/* --------------------------------------------------------------- */
/* sysInitProcIO -- initialize processor base function [startup]   */
/*                                                                 */
/*   returns 0 if OK, -1 if error (but probably cannot recover)    */
/*                                                                 */
/* This must be called before any other sysIO functions; it:       */
/*   -- initializes the processor [SystemInit or equivalent]       */
/*   -- sets the system core clock frequency (processor speed)     */
/*   -- maps the vector table to the start of RAM                  */
/*                                                                 */
/*                                                                 */
/* Assumptions (cases catered for):                                */
/*   -- there is either an external crystal or oscillator, so      */
/*      &chipCrystal will be non-zero                              */
/*                                                                 */
/*   -- speed will be set to highest of 48, 72, or 96 MHz if       */
/*      possible to allow setting of USB 48 MHz                    */
/*      [in future it will make more sense to use PLL1 for USB,    */
/*      but needs some research to ensure this does not exclude    */
/*      devices; also the current approach keeps the code parallel */
/*      with STM32]                                                */
/*                                                                 */
/*   -- chipMaxClock will always be at least twice the external    */
/*      clock                                                      */
/*                                                                 */
/*   -- main oscillator is 2 -> 20 MHz  (could be raised to 25 MHz */
/*      using OSCRANGE bit)                                        */
/* --------------------------------------------------------------- */
// equates missing from library
#define OSCRANGE (1<<4)       // main oscillator range
#define OSCEN    (1<<5)       // main oscillator enabled
#define OSCSTAT  (1<<6)       // main oscillator status

int sysInitProcIO(void) {     // processor initialize (clocks, etc.)
  uint sysclock;              // target clock speed
  uint waits;                 // flash wait time

  // decide on the preferred target system clock speed
  if ((uint)&chipMaxClock>=96000000) sysclock=96000000;
   else if ((uint)&chipMaxClock>=72000000) sysclock=72000000;
   else if ((uint)&chipMaxClock>=48000000) sysclock=48000000;
   else sysclock=(uint)&chipMaxClock;

  /* ------------------------------------------------------------- */
  /* Processor initialization (Reset & Clock Control)              */
  /*                                                               */
  /* This is called after C initialization and the vector table    */
  /* has been copied to RAM.                                       */
  /* ------------------------------------------------------------- */
  // Unfortunately SystemInit() cannot be used because it has a
  // hard-coded fixed oscillator/crystal frequency

  // system reset (in case this is called other than after Reset)
  // select internal oscillator
  LPC_SC->CLKSRCSEL=CLKPWR_CLKSRCSEL_CLKSRC_IRC;
  LPC_SC->SCS=0;                        // disable main oscillator
  stopPLL0();                           // stop PLL0
  LPC_SC->CLKOUTCFG=CLKPWR_CLKOUTCFG_CLKOUTSEL_CPU; // CCLK to CLKOUT
  LPC_SC->PCLKSEL0=0;                   // all peripherals to CCLK/4
  LPC_SC->PCLKSEL1=0;                   // ..

  /* ------------------------------------------------------------- */
  /* Clocks setup ...                                              */
  /* ------------------------------------------------------------- */
  // At this point, the internal oscillator is running; if USEPLL=0,
  // system will use that, otherwise the PLL0 frequency multiplier
  // will be coaxed into action.

  #if USEPLL!=0
  { // declaration block
    uint extclock;                      // external clock speed
    uint pllmult;                       // PLL multiplier (M)
    uint plldivi;                       // PLL divider (N)
    uint plloutd;                       // PLL output divider (D)

    // load external clock speed from link-time constant
    extclock=(uint)&chipCrystal;        // crystal or oscillator

    // start the main oscillator
    // (this seems to be needed whether using a crystal or not)
    LPC_SC->SCS|=OSCEN;                 // enable main oscillator
    while ((LPC_SC->SCS&OSCSTAT)==0);   // wait until Oscillator ready

    LPC_SC->CLKSRCSEL=CLKPWR_CLKSRCSEL_CLKSRC_MAINOSC; // source=main oscillator

    // setPLL0 frequency; the constraints are:
    //   Fin must be in range 32 kHz to 50 MHz
    //   Fcco = (2 * M * extclock / N) must be 275 MHz -> 550 MHz
    // M must be 6->512
    // N must be 1->32
    // the output clock will be Fcco/4  (the 4 in CCLKCFG)
    //
    // Given the target frequencies selected for sysclock, a way to
    // meet these constraints is:
    //
    //   Select Fcco = 288 -- exactly divisible by target frequencies
    //   Select N = 4
    //   Then M = Fcco * 4 / (2 * extclock)
    //
    // M will then be >= 6   so long as extclock <= 96 MHz
    //            and <= 512 so long as extclock >=  2 MHz

    plldivi=4;                               // N
    pllmult=288000000 * 4 / (2 * extclock);  // M
    plloutd=288000000 / sysclock;            // D (output divider)

    LPC_SC->CCLKCFG=plloutd-1;                         // output divider
    LPC_SC->PLL0CFG=((plldivi-1)<<16) | (pllmult-1);   // configure

    sysclock=2*pllmult*extclock/(plldivi*plloutd);     // what will result

    // start up PLL0 as per documented sequence
    LPC_SC->PLL0FEED=0xAA;
    LPC_SC->PLL0FEED=0x55;
    LPC_SC->PLL0CON=CLKPWR_PLL0CON_ENABLE;
    LPC_SC->PLL0FEED=0xAA;
    LPC_SC->PLL0FEED=0x55;
    // wait for PLL lock
    while (!(LPC_SC->PLL0STAT & CLKPWR_PLL0STAT_PLOCK));

    // connect PLL to output and wait for stable
    LPC_SC->PLL0CON=CLKPWR_PLL0CON_ENABLE | CLKPWR_PLL0CON_CONNECT;
    LPC_SC->PLL0FEED  = 0xAA;
    LPC_SC->PLL0FEED  = 0x55;
    // Wait for PLLC0_STAT & PLLE0_STAT (unequated)
    // [NB: NXP code appears to be wrong for this]
    #define STATBITS (CLKPWR_PLL0STAT_PLLE | CLKPWR_PLL0STAT_PLLC)
    while ((LPC_SC->PLL0STAT & STATBITS) != STATBITS);
    } // declaration
  #else
    sysclock=4000000;                   // IRC speed is all you get
  #endif

  // turn peripherals on (probably some do not need to be)
  LPC_SC->PCONP=0x042887DE;             // bits as per example

  // Decide flash wait states (note: if too low, chip won't run)
  //   zero wait state, if 0 < SYSCLK <= 20 MHz
  //   one wait state,  if 20 MHz < SYSCLK <= 40 MHz
  //   two wait states, if 40 MHz < SYSCLK <= 60 MHz
  //   etc.
  if (sysclock<=20000000) waits=0;
   else if (sysclock<=40000000) waits=1;
   else if (sysclock<=60000000) waits=2;
   else if (sysclock<=80000000) waits=3;
   else if (sysclock<=100000000) waits=4;
   else waits=5;                        // 'safe'
  // 3A here always required
  LPC_SC->FLASHCFG=0x3A | (waits<<12);  // configure

  /* ------------------------------------------------------------- */
  /* update SystemCoreClock                                        */
  /* ------------------------------------------------------------- */
  // Again, SystemCoreClockUpdate() cannot be used because the
  // external clock rate is fixed at compile time
  SystemCoreClock=sysclock;

  // debugf("\nM=%d\n", (LPC_SC->PLL0STAT & 0x7FFF) +1);
  // debugf("N=%d\n", ((LPC_SC->PLL0STAT>>16) & 0xFF) +1);
  // debugf("D=%d\n", (LPC_SC->CCLKCFG & 0xFF) +1);

  /* ------------------------------------------------------------- */
  /* set vector table to RAM start and request 16 priority groups  */
  /* ------------------------------------------------------------- */
  SCB->VTOR=(uint32_t)&chipRAMbase & 0x3FFFFF80;
  NVIC_SetPriorityGrouping(3);          // all 16 levels preempt

  return 0;
  } // sysInitProcIO

/* --------------------------------------------------------------- */
/* sysIRQIO -- enable or disable an interrupt                      */
/*                                                                 */
/*   index  -- vector table index number (see sysVector; 0=WDT)    */
/*   enable -- 1 to enable, 0 to disable                           */
/*   level  -- priority level (if enabling) [0..15, 0=high]        */
/*                                                                 */
/*   returns 0 if OK, -1 if index is out of range [-14..max],      */
/*   or -2 if level is too large                                   */
/*                                                                 */
/* This function can be used to change the system vectors, such    */
/* as SysTick, by using a negative index; these 'core' interrupts  */
/* cannot be disabled.                                             */
/* --------------------------------------------------------------- */
// This function should be identical for all CMSIS systems
int  sysIRQIO(int index, flag enable, uint level) {
  // debugf("IRQ index: %d\n", index);
  if (index>=(int)((uint)&chipVectors-16) || index<-14) return -1;

  if (enable) {
    if (level>SYSTEM_MINPRIO) return -2;
    NVIC_SetPriority(index, level);
    if (index>=0) NVIC_EnableIRQ(index);
    }
   else {
    if (index>=0) NVIC_DisableIRQ(index);
    }
  return 0;
  } // sysIRQIO

/* --------------------------------------------------------------- */
/* sysPowerIO -- set a power mode and wait for interrupt           */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_DEEPPOWERDOWN -- power off until Reset or RTC wake     */
/*   SYSTEM_DEEPSLEEP     -- unclocked sleep                       */
/*   SYSTEM_POWERDOWN     -- power down (unclocked, slow start)    */
/*   SYSTEM_SLEEP         -- idle (Wait For Interrupt)             */
/*                                                                 */
/* Currently Tollos only uses SLEEP and DEEPPOWERDOWN.             */
/* --------------------------------------------------------------- */
void sysPowerIO(uint item)  {
  switch(item) {
    case SYSTEM_SLEEP: {       // Sleep (idle)
      CLKPWR_Sleep();
      break;}
    case SYSTEM_DEEPSLEEP: {   // Sleep (unclocked)
      stopPLL0();
      CLKPWR_DeepSleep();
      break;}
    case SYSTEM_POWERDOWN: {   // Power down (unclocked, slow start)
      stopPLL0();
      CLKPWR_PowerDown();
      break;}
    case SYSTEM_DEEPPOWERDOWN:{// Power off until Reset or RTC
      stopPLL0();
      CLKPWR_DeepPowerDown();
      break;}
    default: {                // ignore; return immediately
      break;}
    } // switch
  } // sysPowerIO

/* --------------------------------------------------------------- */
/* sysQueryIO -- query a system item                               */
/*                                                                 */
/*   item -- a constant listed in tollosIO.h                       */
/*                                                                 */
/*   returns current value of the item; negative generally         */
/*     indicates failure                                           */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_CLOCK    -- clock speed, Hz                            */
/*   SYSTEM_RESET    -- 0 if last reset was normal (POR or Reset   */
/*                      pin), 1 if abnormal (Watchdog or Brownout) */
/*   SYSTEM_SYSTICK  -- SysTick interval in us                     */
/*   SYSTEM_WATCHDOG -- Watchdog interval in us                    */
/*                                                                 */
/* --------------------------------------------------------------- */
int sysQueryIO(uint item)  {
  int rc;                     // value to return
  switch(item) {

    case SYSTEM_CLOCK: {      // clock speed, Hz
      rc=SystemCoreClock;     // from the horse's mouth
      break;}

    case SYSTEM_RESET: {      // 0 = normal, 1 = abnormal
      // Watchdog WDTR=bit2 or brownout BODR=bit3 have no equates
      rc=(LPC_SC->RSID & 0x0C)!=0;
      break;}

    case SYSTEM_SYSTICK: {    // current SysTick interval
      rc=systick;
      break;}

    case SYSTEM_WATCHDOG: {   // current Watchdog timeout
      rc=watchtime;
      break;}

    default: {
      rc=-1;                  // bad item
      break;}
    } // switch
  return rc;
  } // sysQueryIO

/* --------------------------------------------------------------- */
/* sysSetIO -- set a system item                                   */
/*                                                                 */
/*   item  -- a constant listed in tollosIO.h                      */
/*   value -- value to set the item to                             */
/*                                                                 */
/*   returns previous value of the item, or a negative number      */
/*     if an error or unknown                                      */
/*                                                                 */
/* Valid items are:                                                */
/*   SYSTEM_SYSTICK  -- SysTick interval in microseconds [>0,      */
/*                      maximum depends on clock rate]             */
/*   SYSTEM_WATCHDOG -- Watchdog interval in microseconds          */
/*                      [>=0, will be rounded up to 256 if less]   */
/* --------------------------------------------------------------- */
int sysSetIO(uint item, int value)  {
  int rc;                // value to return
  switch(item) {

    case SYSTEM_SYSTICK: {              // set SysTick [ms]
      uint prio;                        // SysTick priority
      uint count;                       // work
      rc=systick;

      // check parameter
      if (value<=0) return -1;          // must be at least 1

      // calculate SysTick counter (number to divide clock by); this
      // must be no more than 0xFFFFFF; calculate carefully to
      // avoid multiplication overflow
      count=SystemCoreClock/MILLION * value;
      if (count>0xFFFFFF) return -1;    // SysTick counter out of range

      // The CMSSIS quick setter (also sets to lowest priority, so
      // the original priority must be saved and restored)
      prio=NVIC_GetPriority(SysTick_IRQn);   // save
      if (SysTick_Config(count)) rc=-1;      // failed
       else systick=value;                   // good -- record it
      NVIC_SetPriority(SysTick_IRQn, prio);  // restore

      // The next lines can be used to display the priority, count,
      // and calibration values
      // debugf("SysTick priority: %d\n", prio);
      // debugf("SysTick count: %d\n", count);
      // debugf("SysTick CALIB: %08x\n", SysTick->CALIB);
      break;}

    case SYSTEM_WATCHDOG: {             // watchdog timeout
      uint us=(uint)value;
      if (us>TOLLOS_MAXWATCHDOG*1000 || us<=0) return -1;
      rc=watchtime;
      // for setting the times, microseconds are expected
      // if < 256 then LPC chip will round up to 256
      if (watchtime==0) {                    // first call
        WDT_Init(WDT_CLKSRC_IRC, WDT_MODE_RESET);
        WDT_Start(value);
        }
       else WDT_UpdateTimeOut(value);   // [WDT_Start less Enable]
      watchtime=value;
      break;}

    default: {
      rc=-1;             // bad item
      break;}
    } // switch
  return rc;
  } // sysSetIO


/* --------------------------------------------------------------- */
/* stopPLL0 -- turn off PLL                                        */
/* --------------------------------------------------------------- */
static void stopPLL0(void) {  // turn off PLL0 as per errata
  // needed to stop it continuing to use power
  LPC_SC->PLL0CON &= ~(1<<1); // disconnect PLL0
  LPC_SC->PLL0FEED = 0xAA;    // feed
  LPC_SC->PLL0FEED = 0x55;    // ..
  while ((LPC_SC->PLL0STAT & (1<<25)) != 0x00); // Wait for PLL0 to disconnect
  LPC_SC->PLL0CON &= ~(1<<0); // turn off PLL0
  LPC_SC->PLL0FEED = 0xAA;    // feed
  LPC_SC->PLL0FEED = 0x55;    // ..
  while ((LPC_SC->PLL0STAT & (1<<24)) != 0x00); // Wait for PLL0 to shut down
  } // stopPLL0

