/* --------------------------------------------------------------- */
/* stmI2CIO.c -- I2C device functions for STM                      */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* Low level functions to effect I2C bus I/O.                      */
/*                                                                 */
/* The channels are by default assigned speeds in order to allow   */
/* the use of both slow and faster devices:                        */
/*                                                                 */
/* * channel1: 400 kbit/s                                          */
/* * channel2: 100 kbit/s                                          */
/*                                                                 */
/* These speeds can be changed using i2cInitIO before i2cIO is     */
/* first called for a channel.                                     */
/* --------------------------------------------------------------- */
// Not tested for channel 0 [not accessible on mbed]

#define TOLLOS_IO             // allow use of I/O functions
#include "tollosI2C.h"
#include "stm32f10x.h"
#include "stm32f10x_i2c.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_gpio.h"

// constants
#define I2C_CHANNELS 3        // Total number of channels that might
                              // be present, origin 0
#define I2C_CHANNEL0 0        // 1 if I2C0 accessible, 0 if not
#define I2C_CHANNEL1 1        // 1 if I2C1 accessible, 0 if not
#define I2C_CHANNEL2 1        // 1 if I2C2 accessible

#define I2C_SPEED0      0     // default I2C0 speed [n/a]
#define I2C_SPEED1 400000     // default I2C1 speed
#define I2C_SPEED2 100000     // default I2C2 speed

// statics
static int  businit[I2C_CHANNELS];                // 1 once initialized
static uint busspeed[I2C_CHANNELS]=               // speed of each bus
            {I2C_SPEED0, I2C_SPEED1, I2C_SPEED2}; // [init to defaults]
static I2C_TypeDef * bus[I2C_CHANNELS]=    // STM addresses
            {NULL, I2C1, I2C2};                   //   ..

/* --------------------------------------------------------------- */
/* i2cInitIO -- initialize an I2C channel (hardware interface)     */
/*                                                                 */
/*   channel -- channel number to initialize                       */
/*   hz      -- speed to run the channel                           */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* This function is called automatically, with a default speed     */
/* setting, when i2cIO is called and the channel has not been      */
/* initialized.  Therefore it need only be invoked explicitly for  */
/* a non-default speed.                                            */
/* --------------------------------------------------------------- */
int i2cInitIO(uint channel, uint hz) {
  I2C_InitTypeDef  iinit;               // for configuration
  GPIO_InitTypeDef ginit;               // ..

  switch(channel) {

    #if I2C_CHANNEL1
    case 1: {
      I2C_Cmd(I2C1, DISABLE);           // reset hardware
      I2C_DeInit(I2C1);                 // ..
      // enable I2C1 clock
      RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C1, ENABLE);

      // select PB6 and PB7
      ginit.GPIO_Pin=GPIO_Pin_6 | GPIO_Pin_7;
      break;}
    #endif

    #if I2C_CHANNEL2
    case 2: {
      I2C_Cmd(I2C2, DISABLE);           // reset hardware
      I2C_DeInit(I2C2);                 // ..
      // enable I2C2 clock
      RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C2, ENABLE);

      // select PB10 and PB11
      ginit.GPIO_Pin=GPIO_Pin_10 | GPIO_Pin_11;
      break;}
    #endif

    default: return -1;                 // bad channel
    } // switch

  // set up GPIOs -- NB Alt Function, or could have used bitSet
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOB, ENABLE);
  ginit.GPIO_Speed=GPIO_Speed_50MHz;
  ginit.GPIO_Mode=GPIO_Mode_AF_OD;      // Alt function OpenDrain
  GPIO_Init(GPIOB, &ginit);

  // set up I2C
  iinit.I2C_ClockSpeed=hz;              // up to 400k
  iinit.I2C_Mode=I2C_Mode_I2C;          // required
  iinit.I2C_DutyCycle=I2C_DutyCycle_2;  // fast mode duty cycle
  iinit.I2C_OwnAddress1=0x00;           // initial address not yet known
  iinit.I2C_Ack=I2C_Ack_Disable;        // enabled on I/O
  iinit.I2C_AcknowledgedAddress=I2C_AcknowledgedAddress_7bit; // ..
  // enable I2C peripheral
  I2C_Cmd(bus[channel], ENABLE);
  // apply I2C configuration (after enable)
  I2C_Init(bus[channel], &iinit);
  // [these cannot fail]

  busspeed[channel]=hz;                 // save the speed requested
  businit[channel]=1;                   // now initialized
  return 0;
  } // i2cInitIO

/* --------------------------------------------------------------- */
/* i2cIO -- I2C general I/O function                               */
/*                                                                 */
/*   device -- channel and physical address of device              */
/*             (channel shifted left 16 bits and ORed with 7-bit   */
/*             or 10-bit I2C physical address); currently only     */
/*             7-bit addresses are supported                       */
/*   put    -- start of byte(s) to write                           */
/*   putlen -- count of byte(s) to write [may be 0]                */
/*   get    -- start of byte(s) to read                            */
/*   getlen -- count of byte(s) to read [may be 0]                 */
/*                                                                 */
/*   returns 0 if OK, negative if error                            */
/*                                                                 */
/* put is ignored if putlen is 0 (NULL recommended).               */
/* get is ignored if getlen is 0 (NULL recommended).               */
/*                                                                 */
/* This function will automatically initialize the appropriate     */
/* I2C channel when first called for a channel, unless already     */
/* initialized by an explicit call to i2cInitIO.                   */
/* --------------------------------------------------------------- */

// this should probably be reworked due to errata -- see AN2824 code


int i2cIO(uint device, byte *put, uint putlen, byte *get, uint getlen) {
  uint channel=i2cChannel(device); // channel 0-2
  uint addr=i2cAddr7Bit(device);   // physical address
  I2C_TypeDef *i2c;                // bus shorthand

  if (channel==0 || channel>=I2C_CHANNELS) return -1; // safety check
  i2c=bus[channel];                // shorthand

  // force reset if pending hard error
  #define ITERRS ((I2C_IT_BERR | I2C_IT_AF | I2C_IT_ARLO) &0x00FFFFFF)
  if ((i2c->SR1 & ITERRS)!=0) businit[channel]=0;

  // initialise channel if needed; speed is preserved
  if (!businit[channel]) {
   if (i2cInitIO(channel, busspeed[channel])) return -2;
   }

  // sadly the STM library has only low-level functionality

  // enable acks
  I2C_AcknowledgeConfig(I2C2, ENABLE);
  // wait for bus not busy
  while(I2C_GetFlagStatus(i2c, I2C_FLAG_BUSY));


  // send Start and wait for complete (this sets Master mode)
  I2C_GenerateSTART(i2c, ENABLE);
  while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_MODE_SELECT));

ledSet(0, 1);

  // send device address (shifted) and wait for complete
  I2C_Send7bitAddress(i2c, addr<<1, I2C_Direction_Transmitter);
  while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED));

  // send any data
  for (; putlen>0; putlen--, put++) {
    // send byte and wait for complete
    I2C_SendData(i2c, *put);
    while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_BYTE_TRANSMITTED));

    if (putlen==1 && getlen==0) {  // writing only and one byte to go
      I2C_AcknowledgeConfig(i2c, DISABLE);        // disable Ack
      I2C_GenerateSTOP(i2c, ENABLE);              // request Stop
      }
    }

  // read if requested
  if (getlen>0) {
    // send Start and wait for complete (indicates Read)
    I2C_GenerateSTART(i2c, ENABLE);
    while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_MODE_SELECT));

    // send device address for read and wait for complete
    I2C_Send7bitAddress(i2c, addr<<1, I2C_Direction_Receiver);
    while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED));

    // read the data
    for (; getlen>0; getlen--, get++) {
      if (getlen==1) {                            // sending last byte
        I2C_AcknowledgeConfig(i2c, DISABLE);      // disable Ack
        I2C_GenerateSTOP(i2c, ENABLE);            // request Stop
        }
      // wait for data
      while(!I2C_CheckEvent(i2c, I2C_EVENT_MASTER_BYTE_RECEIVED));
      *get=I2C_ReceiveData(i2c); //Returns Device ID Correctly.
      }
    } // get


  // wait for STOP to be cleared
  while ((i2c->CR1&0x200) == 0x200);    // not an Event :-(

  return 0;  // SUCCESS
  } // i2cIO




