/* --------------------------------------------------------------- */
/* tollos.c -- core code for Tollos                                */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/*                                                                 */
/* This module comprises the core functions for Tollos, including  */
/* system startup, interrupt vector table management, timing, and  */
/* the main scheduler that calls the application ""init()"" and    */
/* ""tick()"" functions.                                           */
/*                                                                 */
/* --------------------------------------------------------------- */
// Naming convention (see also tollos/code_conventions.html):
//
//   sys* -- documented functions and data in the Tollos 'API'
//   tos* -- private functions and data (may be redefined)
//
// After the statics the functions are in alphabetic order.

#define  TOLLOS_IO                 // allow use of I/O functions
#include "tollos.h"

// Stack instrumentation constants [stack usage is detected by
// placing footprints in RAM every so often; their being
// overwritten allows an estimate of the stack low water mark]
#define MAGIC    0x4D46432E        // unlikely sequence
#define MAGICGAP 32                // footprint gap

// Built-in interrupt handlers and utility
static uint tosGetIPSR(void);
static void tosHandleSysTick(void);
static void tosHandleUnexpected(void);

// ---------------------------------------------------------------
// Static interrupt vector table for Tollos on ARM processors
//
// The table comprises the basic ARM-defined 16 vectors, of which
// the first two are 'special':
//   0: the initial stack pointer (address, not a function pointer)
//   1: the address of the startup function [hard or soft reset]
//
// Vectors are by default mapped to sysHandleUnexpected, which
// issues a debug message then powers off.
//
// SysTick is handled by Tollos, so its handler can be statically
// initialized here.
//
// This static table must appear at the base of ROM (using the
// "vectors" section name) for startup; it is copied to RAM during
// startup and the following chip-dependent vectors are set to
// tosHandleUnexpected.
//
// External interrupts are device-dependent and hence their mappings
// and handlers must be provided by the system-dependent code,
// The number of vectors is set in the link script, and vectors
// are added to the vector table in RAM by calling sysVector.
//
// This table should be forced to be included using
//   KEEP(*(vectors))
// in the link script; since the code entrypoint is in this table
// this should cause all the code to be included, too (if used)
// and avoids needing to specify the name of the entrypoint in the
// link script.  [However, some versions of ld seem not to do
// this.]
// ---------------------------------------------------------------
// count of basic vectors
#define BVCOUNT (sizeof(basicvectors)/sizeof(vector))
#define UNEX    tosHandleUnexpected     // shorthand
#define OFFSET_SYSTICK -1               // offset index of SysTick vector
__attribute__ ((section("vectors")))    // used to force load point
static basicvectors tosBasicVectors = {
  (void *) &chipRAMend,                 //  0: stack pointer
  (vector) sysStartup,                  //  1: entry point to start
  (vector) UNEX,                        //  2: NMI
  (vector) UNEX,                        //  3: HardFault
  (vector) UNEX,                        //  4: MemManage
  (vector) UNEX,                        //  5: BusFault
  (vector) UNEX,                        //  6: UsageFault
  (vector) 0,                           //  7: (reserved)
  (vector) 0,                           //  8: (reserved)
  (vector) 0,                           //  9: (reserved)
  (vector) 0,                           // 10: (reserved)
  (vector) UNEX,                        // 11: SVC
  (vector) UNEX,                        // 12: DebugMonitor
  (vector) 0,                           // 13: (reserved)
  (vector) UNEX,                        // 14: PendSVC
  (vector) tosHandleSysTick             // 15: SysTick [offset -1]
  };

// ---------------------------------------------------------------
// Other statics (statics are always private in Tollos)
// ---------------------------------------------------------------
// sysTick time in microseconds [must be >0, and a factor of MILLION
// so that there will be an integer number of ticks in a second]
#define TOS_SYSTICK 100
#if ((MILLION/TOS_SYSTICK)*TOS_SYSTICK)!=MILLION
  #error TOS_SYSTICK is not a factor of 1000000
#endif

// tick timing and statistics, in microseconds
static uint tosTickTarget=TOLLOS_TICKTIME; // target micoseconds per tick
static uint tosTickLast=0;              // time used for last tick() call
static uint tosTickMax=0;               // maximum time used for a tick()

// lookaside for stack accounting
static char *tosLastseen;               // highest MAGIC

// The following are changed in an interrupt handler, so cannot be cached
static volatile uint tosMicros=0;       // microseconds since start
static volatile uint tosSeconds=0;      // seconds since start
// [Micros and seconds are split (rather than using int64) because it
// is often useful to have that split and be able to determine it
// without a division being necessary.  It also keeps this core code
// as 'native' 32-bit.]

/* --------------------------------------------------------------- */
/* sysPause -- delay for a number of milliseconds                  */
/*                                                                 */
/*   ms -- milliseconds to pause for (may be 0)                    */
/*                                                                 */
/* If ms is 0, sysPause returns immediately.  Otherwise, it waits  */
/* for an interval of at least ms-1 milliseconds and no more than  */
/* ms milliseconds.  If possible the processor will sleep while    */
/* waiting (interrupts will be handled).                           */
/*                                                                 */
/* sysPause may be used in init() or later, but note that it does  */
/* not extend the tick period, so if called within tick() the      */
/* pause time should be less than the tick period.                 */
/* --------------------------------------------------------------- */
// The 'no more than' constraint assumes that sysTick is <= 1 ms
#if TOS_SYSTICK>1000
  #error TOS_SYSTICK is > 1000, rework of sysPause needed
#endif
void sysPause(uint ms) {
  uint endsecs, endmicros;         // target pause end time

  sysTime(&endsecs, &endmicros);        // now
  endmicros+=ms*1000;                   // calculate end time
  if (endmicros>=MILLION) {             // carry ..
    endsecs++;
    endmicros-=MILLION;
    }

  // wait until the end time [see commentary in sysStartup]
  for (;;) {
    uint thissecs, thismicros;           // work
    sysTime(&thissecs, &thismicros);     // now
    if (thissecs>endsecs) break;         // definitely past the time
    if (thissecs==endsecs && thismicros>=endmicros) break;
    sysPowerIO(SYSTEM_SLEEP);
    } // for
  } // sysPause

/* --------------------------------------------------------------- */
/* sysQuery -- query a system item                                 */
/*                                                                 */
/*   item -- a constant listed in tollos.h                         */
/*                                                                 */
/*   returns current value of the item, or -1 if the item is       */
/*     unknown                                                     */
/*                                                                 */
/* Valid items are all items valid for sysSet, and also:           */
/*   SYS_CLOCK     -- processor clock speed, Hz                    */
/*   SYS_RESET     -- cause of last reset: 0=normal (POR or Reset  */
/*                    pin), 1=unusual (Watchdog or Brownout)       */
/*   SYS_STACKSIZE -- total RAM available for stack, bytes         */
/*   SYS_STACKUSED -- maximum stack used since startup, bytes      */
/*   SYS_TICKLAST  -- time used for last tick() call, microseconds */
/*                    (maximum 1000s)                              */
/* --------------------------------------------------------------- */
int sysQuery(uint item)  {
  int rc;                // value to return
  switch(item) {

    case SYS_CLOCK: {    // clock speed, Hz
      return sysQueryIO(SYSTEM_CLOCK);
      break;}

    case SYS_RESET: {    // reset cause
      return sysQueryIO(SYSTEM_RESET);
      break;}

    case SYS_STACKSIZE: {// total RAM available for stack
      rc=&chipRAMend-&chipRAMtop;
      break;}

    case SYS_STACKUSED: {// low water mark; maximum stack used
      char *target;                          // work
      rc=-1;                                 // in case no magic found
      for (target=tosLastseen; target>&chipRAMtop; target-=MAGICGAP) {
        if (*(int *)target==MAGIC) {         // untrodden magic
          tosLastseen=target;
          rc=&chipRAMend-target-sizeof(int); // calculate used
          break;                             // out of the for
          }
        } // for
      break;}

    case SYS_TICKLAST: { // latest tick() time, us
      rc=tosTickLast;
      break;}

    case SYS_TICKMAX: {  // maximum tick() time, us
      rc=tosTickMax;
      break;}

    case SYS_TICKTIME: { // current tick target, us
      rc=tosTickTarget;
      break;}

    case SYS_WATCHDOG: { // watchdog time, ms
      int watchtime=sysQueryIO(SYSTEM_WATCHDOG);
      if (watchtime>0) watchtime/=1000; // to ms, if good
      return watchtime;
      break;}

    default: {
      rc=-1;             // bad item
      break;}
    } // switch

  return rc;
  } // sysQuery

/* --------------------------------------------------------------- */
/* sysSet -- set a system item                                     */
/*                                                                 */
/*   item  -- a constant listed in tollos.h                        */
/*   value -- value to set the item to                             */
/*                                                                 */
/*   returns previous value of the item, or a negative number      */
/*     if the item is unknown or a bad value is given              */
/*                                                                 */
/* Valid items are:                                                */
/*   SYS_TICKMAX   -- maximum recorded tick() time, microseconds   */
/*                    [0<=value]                                   */
/*   SYS_TICKTIME  -- tick() call interval, microseconds [0<value] */
/*   SYS_WATCHDOG  -- watchdog interval, milliseconds              */
/*                    [0<value<=10s]                               */
/*                                                                 */
/* SYS_TICKMAX can be set to any non-negative value, but typically */
/* only a value of 0 is used (to reset the 'high water mark').     */
/* The maximum value for SYS_TICKMAX is clamped to 1000 seconds.   */
/*                                                                 */
/* SYS_TICKTIME should be a multiple of the internal SysTick       */
/* (currently 100us) for accurate tick timing.                     */
/*                                                                 */
/* Depending on the processor, SYS_WATCHDOG may be rounded up,     */
/* possibly to a multiple of 4ms.                                  */
/* --------------------------------------------------------------- */
int sysSet(uint item, int value)  {
  int rc;                // value to return
  switch(item) {

    case SYS_TICKMAX: {  // maximum tick() time, us
      if (value>=0) {    // any non-negative value
        rc=tosTickMax;
        tosTickMax=min(value, BILLION);
        }
       else rc=-1;
      break;}

    case SYS_TICKTIME: { // tick time, us [>0]
      if (value>0) {     // must be positive
        rc=tosTickTarget;
        tosTickTarget=value;
        }
       else rc=-1;
      break;}

    case SYS_WATCHDOG: { // watchdog time, ms
      // 0 < value <= n000
      if (value<=0 || value>TOLLOS_MAXWATCHDOG) return -1;
      return sysSetIO(SYSTEM_WATCHDOG, value*1000);
      break;}

    default: {
      rc=-1;             // bad item
      break;}
    } // switch
  return rc;
  } // sysSet

/* --------------------------------------------------------------- */
/* sysStartup -- first function called after reset                 */
/*                                                                 */
/* This function sets up the hardware and runs the application     */
/* as follows:                                                     */
/*                                                                 */
/* #initialize C statics                                           */
/* #initialize Interrupt Vector table in RAM                       */
/* #initialize stack footprints (used for reporting stack usage)   */
/* #initialize processor clocks, PLL, etc.                         */
/* #set up the default tick time (40 ms = 25 Hz)                   */
/* #invoke the application init() function (if any)                */
/* #thereafter invoke the application tick() function (if any) at  */
/* the appropriate interval.                                       */
/*                                                                 */
/* This could be called explicitly to restart the processor and    */
/* software, but note that peripherals will not receive a hardware */
/* reset.                                                          */
/* --------------------------------------------------------------- */
// Declaration blocks are used here to avoid never-reused stack
void sysStartup(void) {
  int rc;                          // general use returncode
  // [if add more to the stack, check the generated asm -- may need
  // to change the position of the first MAGIC marker below]

  /* ------------------------------------------------------------- */
  /* C initialization                                              */
  /* ------------------------------------------------------------- */
  { // declaration block
    char *source, *target;         // work
    // Initialize statics (.data section) by copy from ROM
    source=&chipDataSource;
    for (target=&chipDataStart; target<&chipDataEnd;) *target++ = *source++;

    // Clear uninitialized statics (.bss section)
    for (target=&chipZeroStart; target<&chipZeroEnd;) *target++ = 0;
    } // declaration

  /* ------------------------------------------------------------- */
  /* Copy the basic vector table to RAM                            */
  /* ------------------------------------------------------------- */
  // This piece of code assumes ARM:
  //   sizeof(int) = sizeof(int *) = sizeof(function pointer)
  { // declaration block
    int *vsource, *vtarget;        // work
    vtarget=(int *)&chipVectorStart; // in RAM
    for (vsource=(int *)&tosBasicVectors;
         vsource<(int *)&tosBasicVectors+BVCOUNT;) *vtarget++ = *vsource++;
    // fill in remaining vectors with pointer to handler
    for(; vtarget<(int *)&chipVectorEnd;)
      *vtarget++ = (int)tosHandleUnexpected;
    } // declaration

  /* ------------------------------------------------------------- */
  /* Stack footprints -- before any interrupts -- every GAP bytes  */
  /* ------------------------------------------------------------- */
  // Note that some stack will already be in use.
  // Currently stack usage detection is top-down for speed: this
  // could fail if MAGIC just happened to be written into an
  // inspected stack slot.  This should probably change to search
  // bottom-up, after some timing masurements
  { // declaration block
    char *target;                  // work
    tosLastseen=&chipRAMend-MAGICGAP*2; // save top spot as lookaside
    for (target=tosLastseen; target>&chipRAMtop; target-=MAGICGAP) {
      *(int *)target=MAGIC;
      } // for
    } // declaration

  /* ------------------------------------------------------------- */
  /* Processor and Vector table initialization                     */
  /* ------------------------------------------------------------- */
  rc=sysInitProcIO();         // processor initialize (clocks, etc.)
                              // also sets to use vectors in RAM
  if (rc!=0) debugf("sysInitProc rc=%d\n", rc); // optimistic ...

  /* ------------------------------------------------------------- */
  /* Invoke application initialization function (if it exists)     */
  /*                                                               */
  /* The init function should return:                              */
  /*   0 -- continue on to start calling tick()                    */
  /*   1 -- halt (powerdown) without calling tick()                */
  /* ------------------------------------------------------------- */
  // Start the SysTick first, so sysPause can be used in init()
  rc=sysIRQIO(OFFSET_SYSTICK, 1, SYSTEM_TICKPRIO); // set priority
  if (rc!=0) debugf("SysTick Prio rc=%d\n", rc);
  rc=sysSetIO(SYSTEM_SYSTICK, TOS_SYSTICK);        // set microseconds
  if (rc!=0) debugf("SysTick Set rc=%d\n", rc);

  rc=init ? init() : 0;            // call init() if exists; else rc=0

  /* ------------------------------------------------------------- */
  /* Invoke application tick function on each target count (if it  */
  /* exists).                                                      */
  /*                                                               */
  /* The tick function should return:                              */
  /*   0 -- continue calling tick() at the designated interval     */
  /*   1 -- halt (powerdown)                                       */
  /*                                                               */
  /* Note that overrun is possible if the tick function takes too  */
  /* long, however ticks will self-correct over time because a     */
  /* synchronous (absolute) trigger for tics is used.              */
  /* ------------------------------------------------------------- */
  // Earlier (1.24 or earlier) versions did not have an absolute tick
  // counter and used a sysTick counter; this opened a tiny window of
  // about one sysTick in one million being 'lost'.

  if (rc==0 && tick) {             // init was clean, and tick exists
    uint lastsecs, lastmicros;     // noted start time for this tick
    uint nextsecs, nextmicros;     // target start time for next tick
    uint thissecs, thismicros;     // this instant (now)
    int  secs, micros;             // for elapsed time calculation

    sysTime(&nextsecs, &nextmicros);    // now
    for (;;) {                          // tick() forever
      lastsecs=nextsecs;                // save ..
      lastmicros=nextmicros;            // ..
      nextmicros+=tosTickTarget;        // set next target time
      if (nextmicros>=MILLION) {        // carry ..
        nextsecs++;
        nextmicros-=MILLION;
        }
      sysFeedDogIO();                   // gratify the pekingese
      rc=tick();                        // call user function
      if (rc!=0) break;                 // powerdown request

      // calculate time elapsed; if this is more than 1000 seconds
      // result is clamped to 1000 seconds
      sysTime(&thissecs, &thismicros);
      micros=(int)thismicros-(int)lastmicros;     // could be -ve
      secs=thissecs-lastsecs;                     // always >=0
      if (secs>2000) secs=2000;                   // force clamp
      micros+=secs*MILLION;                       // add in seconds
      if (micros>BILLION) micros=BILLION;         // clamp

      tosTickLast=micros;               // save how long that took
      if (tosTickLast>tosTickMax) tosTickMax=tosTickLast; // note max

      // wait until the next tick target time; this sleeps until the
      // next interrupt, if allowed, to minimise power consumption
      // [on some systems this cannot be done (because it may hang
      // remote debug, for example), in which case the wait becomes a
      // hard loop]
      for (;;) {
        if (thissecs>nextsecs) break;   // definitely past the time
        // here, thissecs<=nextsecs; reached the time if secs are
        // equal and micros equal or greater
        if (thissecs==nextsecs && thismicros>=nextmicros) break;
        sysPowerIO(SYSTEM_SLEEP);
        sysTime(&thissecs, &thismicros); // get new 'now'
        } // for

      } // tick loop
   } // init was OK

  /* any drop through to here means "halt" */
  debugf("Halted\n");
  sysPause(250);                    // allow time to send message

  // later this could use rc for other power down modes
  sysPowerIO(SYSTEM_DEEPPOWERDOWN); // until Reset or RTC
  } // sysStartup


/* --------------------------------------------------------------- */
/* sysTime -- return timestamp                                     */
/*                                                                 */
/*   seconds -- receives seconds since Tollos started              */
/*   micros  -- receives microseconds in current second            */
/*                                                                 */
/* Seconds and micros are set to the number of seconds and         */
/* microseconds since the Tollos was started, rounded to the       */
/* latest internal Tollos sysTick.  The interval between internal  */
/* sysTicks is 1000 microseconds (1ms), so micros will be a        */
/* multiple of 1000.                                               */
/*                                                                 */
/* The seconds counter will wrap approximately every 136.1 years.  */
/* --------------------------------------------------------------- */
// 365.2425 days/year
void sysTime(uint *seconds, uint *micros) {
  // a microseconds value is only valid if seconds values read
  // before and after it are identical
  for (;;) {
    uint oldsecs=tosSeconds;
    *micros=tosMicros;
    *seconds=tosSeconds;
    if (*seconds==oldsecs) break;  // almost always immediate
    }
  return;
  } // sysTime

/* --------------------------------------------------------------- */
/* sysVector -- set an interrupt vector address                    */
/*                                                                 */
/*   index -- item # in vectors list (e.g., Watchdog is 0)         */
/*   addr  -- address to set, or NULL to cancel handler            */
/*   level -- priority level; 0=highest, 15=lowest [all preempt]   */
/*                                                                 */
/*   returns 0 if the handler was changed to the new handle, or    */
/*     cancelled, successfully; 1 if the handler was already set   */
/*     as requested; -1 if the vector is already in use with a     */
/*     different handler and addr is not NULL; -2 if item is too   */
/*     large; -3 if the priority level is too large                */
/*                                                                 */
/* Note that index is the vector number excluding the first 16     */
/* slots.                                                          */
/*                                                                 */
/* Setting an address automatically enables interrupts for that    */
/* item, and cancelling it (with NULL) disables interrupts.        */
/*                                                                 */
/* When enabling, the priority level of the interrupt is also set; */
/* a higher priority interrupt (smaller value for level) will      */
/* preempt a lower priority.                                       */
/*                                                                 */
/* The error checking here protects against attempted use of the   */
/* same vector by two different handlers.  It is not an error to   */
/* replace a handler by itself (hence it is safe to always set a   */
/* handler when needed), and the return code will indicate         */
/* whether this was the first time it was set.                     */
/* --------------------------------------------------------------- */
// This could easily be extended to allow setting the 'system' slots
// by allowing index to take the values -14 through -1 (i.e., an int)
int sysVector(uint index, vector addr, uint level) {
  vector *target;                  // vector to change

  if (index>=(uint)&chipVectors-16) return -2;

  target=(vector *)(&chipVectorStart+(index+16)*4); // select vector

  if (*target==addr) return 1;     // already set as required

  if (addr!=NULL && *target!=&tosHandleUnexpected) return -1;

  if (addr==NULL) *target=&tosHandleUnexpected;
   else *target=addr;

  if (sysIRQIO(index, addr!=NULL, level)<0) return -3;
  return 0;
  } // sysVector

/* --------------------------------------------------------------- */
/* tosGetIPSR -- get Interrupt Program Status Register             */
/*                                                                 */
/* This directly reads the hardware register to report the cause   */
/* of an interrrupt; only used after an unexpected interrupt.      */
/* --------------------------------------------------------------- */
static uint tosGetIPSR(void) {
  __asm("mrs r0, ipsr");           // move special register to R0 ..
  __asm("bx lr");                  // .. and return
  return 0;                        // [placate compiler; not reached]
  } // tosGetIPSR

/* --------------------------------------------------------------- */
/* tosHandleSysTick -- SysTick interrupt handler                   */
/*                                                                 */
/* This simply bumps the timestamp values and returns ASAP.        */
/* --------------------------------------------------------------- */
// tosGetIPSR() returns 15 (0x0f) in this handler
static void tosHandleSysTick(void) {
  tosMicros+=TOS_SYSTICK;          // bump microseconds
  if (tosMicros==MILLION) {        // new second
    tosSeconds++;
    tosMicros=0;
    }

  // for testing, uncomment next line; specified pin should then
  // change state every TOS_SYSTICK microseconds
  // bitSet(PORT0, BIT26, BIT_INVERT);  // mbed p18
  // bitSet(PORTA, BIT2,  BIT_INVERT);  // primer2 CX-USART-TX p16
  } // tosHandleSysTick

/* --------------------------------------------------------------- */
/* tosHandleUnexpected -- Unexpected interrupts handler            */
/*                                                                 */
/* This reports any unexpected interrupt and then halts.           */
/* --------------------------------------------------------------- */
static void tosHandleUnexpected(void) {
  debugf("UnExp Halt %d\n", tosGetIPSR());   // show which interrupt
  sysPause(250);                             // allow time to send message
  sysPowerIO(SYSTEM_DEEPPOWERDOWN);          // until Reset or RTC
  } // tosHandleUnexpected

