/* --------------------------------------------------------------- */
/* tollosAcc.c -- accelerometer functions for Tollos               */
/* --------------------------------------------------------------- */
/*                                                                 */
/* Copyright (c) Mike Cowlishaw, 2010-2012.  All rights reserved.  */
/*                                                                 */
/* Permission to use, copy, modify, and/or distribute this         */
/* software for any purpose with or without fee is hereby granted, */
/* provided that the above copyright notice and this permission    */
/* notice appear in all copies, and that notice and the date of    */
/* any modifications be added to the software.                     */
/*                                                                 */
/* This software is provided "as is".  No warranties, whether      */
/* express, implied, or statutory, including, but not limited to,  */
/* implied warranties of merchantability and fitness for a         */
/* particular purpose apply to this software.  The author shall    */
/* not, in any circumstances, be liable for special, incidental,   */
/* or consequential damages, for any reason whatsoever.            */
/*                                                                 */
/* --------------------------------------------------------------- */
/* These functions provide basic accelerometer functions for       */
/* Tollos, intended to be device-independent (although at the      */
/* moment an I2C connection is assumed).                           */
/*                                                                 */
/* To use, the device must be initialized using accInit.  This     */
/* indicates which I2C channel the device is on and which port/pin */
/* is used for DataReady interrupt.  By default the accelerometer  */
/* range will be at least 2g; this may be changed with accRange.   */
/*                                                                 */
/* The latest accelerometer reading can be read safely using       */
/* accGetXYZ, which returns the three axes in milli-g.  The axes   */
/* directions are as defined by the device manufacturer.           */
/*                                                                 */
/* The default update rate is selected by the device driver code   */
/* and will be at least 25Hz (that is, the data will be no more    */
/* than 40ms old).  Update is not necessarily synchronous with the */
/* Tollos tick rate.   The rate can be changed with accRate.       */
/*                                                                 */
/* tollosAcc will buffer up to four accelerometer readings.        */
/* This allows an update rate of double the application's          */
/* wake-up frequency, plus one missed wakeup.  An application      */
/* should therefore call accGetXYZ repeatedly until 'no data       */
/* available' is reported in order to clear queued data and to     */
/* avoid overrun.                                                  */
/*                                                                 */
/* --------------------------------------------------------------- */

#define  TOLLOS_IO            // allow use of I/O functions
#include "tollos.h"

// vector of readings from Accelerometer
typedef struct {
  int x;                      // X reading [mg]
  int y;                      // Y reading [mg]
  int z;                      // Z reading [mg]
  } accData;

// Ring-buffer of readings; this is used conventionally to avoid race
// conditions between the interrupt handler and the accGetXYZ
// function.
//
// In particular:
//
//   * The write pointer (index) is only incremented (and wrapped if
//     need be) altered by the interrupt handler, after good data
//     have been read.
//
//   * The read pointer is only incremented (and wrapped if need be)
//     after the Get function has copied data to the caller.
//
//   * If the read and write pointers are equal, the ring is empty.
//
//   * If stepping the write pointer would make the write pointer
//     equal the read pointer then an overrun has occurred and the
//     write pointer is not stepped.  This is reported (and cleared)
//     at the next read.
//
// Hence: the two levels of code always write to and read from
// different entries in the ring, and each level independently
// modifies the pointer that it 'owns'.
//
// The default size of the buffer ring allows for four entries.
#define RINGSIZE 4

static accData accRing[RINGSIZE+1];     // +1 for the working slot
static int     accWrite=0;              // write index
static int     accRead=0;               // read index

// Status flags
static flag    accReady=0;              // accInit has been successful
static flag    accOverrun=0;            // overrun has occurred

// local function
static void accHandler(void); // interrupt handler

/* --------------------------------------------------------------- */
/* accGetXYZ -- read latest x, y, and z accelerations              */
/*                                                                 */
/*   x, y, z -- receive accelerations [mg: 1000 = 1.000 g], or are */
/*     unchanged if no data are available.                         */
/*                                                                 */
/*   returns 0 if data good, +1 if data is good but an overrun has */
/*     occurred since the last read, -1 if no good data from read, */
/*     or -2 if the device is not initialized.                     */
/*                                                                 */
/* Results are returned scaled to milli-gs (1000 = 1g), without    */
/* any calibration adjustments.                                    */
/*                                                                 */
/* When +1 (overrun) is returned the overrun status is cleared and */
/* future good reads will return 0 until another overrun occurs.   */
/*                                                                 */
/* This must not be called from an interrupt level with higher     */
/* priority than the accelerometer interrupt.                      */
/* --------------------------------------------------------------- */
int accGetXYZ(int *x, int *y, int*z) {
  // see above for the procedure used here

  if (!accReady) return -2;             // unready
  if (accWrite==accRead) return -1;     // ring empty

  // return data from the 'read' slot
  *x=accRing[accRead].x;
  *y=accRing[accRead].y;
  *z=accRing[accRead].z;

  accRead++;                            // bump index
  if (accRead==RINGSIZE+1) accRead=0;   // wrap
  if (!accOverrun) return 0;            // is good
  accOverrun=0;                         // had overrun; clear flag
  return 1;                             // .. and report
  } // accGetXYZ

/* --------------------------------------------------------------- */
/* accHandler -- interrupt handler, called when Data Ready         */
/*                                                                 */
/* This reads the data into the 'write' slot, then when and if     */
/* bumps the write index unless this would overrun.                */
/* --------------------------------------------------------------- */
static void accHandler(void) {
  int rc;                          // work
  int newWrite;                    // tentative write index

  rc=accGetXYZIO(&accRing[accWrite].x, &accRing[accWrite].y,
                 &accRing[accWrite].z);
  // interrupt is cleared by the read
  if (rc<0) return;                          // bad data

  newWrite=accWrite+1;                       // bump index
  if (newWrite==RINGSIZE+1) newWrite=0;      // wrap

  if (newWrite!=accRead) accWrite=newWrite;  // OK to use it
   else accOverrun=1;                        // [don't bump]
  } // accHandler

/* --------------------------------------------------------------- */
/* accInit -- configure and initialize accelerometer               */
/*                                                                 */
/*   channel -- I2C channel on which device is to be found         */
/*   drport  -- port for the Data Ready interrupt pin              */
/*   drpin   -- pin number (not bit mask) for the Data Ready       */
/*              interrupt pin                                      */
/*                                                                 */
/*   returns 0 if initialized OK, negative if error                */
/*                                                                 */
/* This sets up the pin and interrupt handler for the Data Ready   */
/* interrupt from the device then calls the device driver          */
/* initialization function to start the device.                    */
/*                                                                 */
/* The accInitIO function will set:                                */
/* * an update rate of at least 25 Hz (but not necessarily higher  */
/* than that, to save power); call accRate if a different rate is  */
/* needed                                                          */
/* * a range of at least 2g; call accRange if a larger range is    */
/* needed                                                          */
/* * an interrupt on Data Ready (this is handled by tollosAcc).    */
/*                                                                 */
/* Note that it can be 40ms or more before the first valid data    */
/* are available, so an accGetXYX call before that time may report */
/* no data available, as might a call that reads the device less   */
/* than 25ms after the last read.                                  */
/* --------------------------------------------------------------- */
int accInit(uint channel, uint drport, uint drpin) {
  int rc;                     // work

  accWrite=0;                 // clear indexes ..
  accRead=0;                  // ..

  // set up interrupt pin from number (bitFuns use a mask)
  rc=bitEnable(drport, 1<<drpin, BIT_RISE, accHandler);
  if (rc==0) rc=accInitIO(channel);
  if (rc==0) accReady=1;     // record good status
  return rc;
  } // accInit

/* --------------------------------------------------------------- */
/* accRange -- change range of accelerometer                       */
/*                                                                 */
/*   range -- range requested in milli-g [1000 = 1 g]              */
/*                                                                 */
/*   returns the chosen range if changed OK, -1 if error (range    */
/*     too high)                                                   */
/*                                                                 */
/* The accelerometer will be set to a smallest available range     */
/* that is >= the requested range, to maximize the accuracy of     */
/* returned results.                                               */
/* --------------------------------------------------------------- */
int accRange(uint range) {
  return accRangeIO(range);   // pass on and back
  } // accRange

/* --------------------------------------------------------------- */
/* accRate -- change update rate of accelerometer                  */
/*                                                                 */
/*   rate -- rate requested [Hz, 1..200]                           */
/*                                                                 */
/*   returns the chosen rate if changed OK, -1 if error (rate zero */
/*     or too high)                                                */
/*                                                                 */
/* The accelerometer will be set to a slowest available rate       */
/* that is >= the requested rate, to minimize resource usage.      */
/*                                                                 */
/* To ensure working with 100kHz I2C, the maximum rate is limited  */
/* to 200Hz -- but note that this may use significant processor    */
/* bandwidth.                                                      */
/* --------------------------------------------------------------- */
int accRate(uint rate) {
  return accRateIO(rate);     // pass on and back
  } // accRate

/* --------------------------------------------------------------- */
/* accStop -- stop and power-down accelerometer                    */
/*                                                                 */
/*   returns 0 if accelerometer stopped OK, negative if an error   */
/*     or it was not running.                                      */
/* --------------------------------------------------------------- */
int accStop(void) {
  int rc;                     // work
  accReady=0;                 // invalidate any data
  rc=accStopIO();             // stop the device
  accRead=0;                  // be clean
  accWrite=0;                 // ..
  return rc;
  } // accStop

